import torch
from transformers import AutoModelForCausalLM, AutoProcessor

from ..model_registry import ModelInputOutput, register_model
from ..models import BaseModel, track_inference


@register_model(
    "microsoft/Florence-2-large", "transformers", ModelInputOutput.IMAGE_TEXT_TO_TEXT
)
@register_model(
    "microsoft/Florence-2-base",
    "transformers",
    ModelInputOutput.IMAGE_TEXT_TO_TEXT,
)
@register_model(
    "microsoft/Florence-2-large-ft",
    "transformers",
    ModelInputOutput.IMAGE_TEXT_TO_TEXT,
)
@register_model(
    "microsoft/Florence-2-base-ft",
    "transformers",
    ModelInputOutput.IMAGE_TEXT_TO_TEXT,
)
class Florence2(BaseModel):
    def __init__(
        self,
        model_id: str,
        device: str = "cpu",
        dtype: str = "float32",
    ):
        super().__init__(model_id, device, dtype)
        self.load_model()

    def load_model(self):
        self.model = AutoModelForCausalLM.from_pretrained(
            self.model_id, trust_remote_code=True
        ).to(self.device, self.dtype)
        self.model.eval()
        self.model = torch.compile(self.model, mode="max-autotune")
        self.processor = AutoProcessor.from_pretrained(
            self.model_id, trust_remote_code=True
        )

    @track_inference
    def infer(self, image: str, prompt: str = None, **generate_kwargs) -> str:
        output = self.infer_batch([image], [prompt], **generate_kwargs)
        return output[0]

    @track_inference
    def infer_batch(
        self, images: list[str], prompts: list[str] = None, **generate_kwargs
    ) -> list[str]:
        images = self.parse_images(images)
        inputs = self.processor(text=prompts, images=images, return_tensors="pt").to(
            self.device, self.dtype
        )

        if "max_new_tokens" not in generate_kwargs:
            generate_kwargs["max_new_tokens"] = 1024
        if "num_beams" not in generate_kwargs:
            generate_kwargs["num_beams"] = 3

        with torch.inference_mode():
            generated_ids = self.model.generate(
                input_ids=inputs["input_ids"],
                pixel_values=inputs["pixel_values"],
                **generate_kwargs,
            )

        generated_text = self.processor.batch_decode(
            generated_ids, skip_special_tokens=False
        )

        parsed_answers = [
            self.processor.post_process_generation(
                text, task=prompt, image_size=(img.width, img.height)
            ).get(prompt)
            for text, prompt, img in zip(generated_text, prompts, images)
        ]

        return parsed_answers



import pytest
import torch

from xinfer.models import BaseModel


class MockModel(BaseModel):
    def load_model(self):
        pass

    def infer(self, image: str, prompt: str):
        pass

    def infer_batch(self, images: list[str], prompts: list[str]):
        pass


@pytest.fixture
def base_model():
    return MockModel("test_model", "cpu", "float32")


def test_base_model_init(base_model):
    assert base_model.model_id == "test_model"
    assert base_model.device == "cpu"
    assert base_model.dtype == torch.float32
    assert base_model.num_inferences == 0
    assert base_model.total_inference_time == 0.0
    assert base_model.average_latency == 0.0



"""Top-level package for xinfer."""

__author__ = """Dickson Neoh"""
__email__ = "dickson.neoh@gmail.com"
__version__ = "0.1.2"

from .core import create_model, list_models
from .model_registry import ModelInputOutput, register_model
from .models import BaseModel
from .utils import (
    ollama_available,
    timm_available,
    transformers_available,
    ultralytics_available,
    vllm_available,
)

if timm_available:
    from .timm import *
if transformers_available:
    from .transformers import *
if ultralytics_available:
    from .ultralytics import *
if vllm_available:
    from .vllm import *
if ollama_available:
    from .ollama import *

from .viz import launch_gradio_demo

__all__ = [
    "create_model",
    "list_models",
    "register_model",
    "BaseModel",
    "ModelInputOutput",
    "launch_gradio_demo",
]



import functools
import time
from abc import ABC, abstractmethod

import requests
import torch
from loguru import logger
from PIL import Image
from rich import box
from rich.console import Console
from rich.table import Table


def track_inference(func):
    """
    Decorator to track the inference time of a model. Put this on the inference methods of a model.

    Example:
        @track_inference
        def infer(self, image: str, prompt: str):
            ...
            return result

        @track_inference
        def infer_batch(self, images: list[str], prompts: list[str]):
            ...
            return result
    """

    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):
        start_time = time.perf_counter()
        result = func(self, *args, **kwargs)
        end_time = time.perf_counter()

        inference_time = (end_time - start_time) * 1000
        self.total_inference_time += inference_time

        if func.__name__ == "infer_batch":
            # For batch inference, increment by the number of images
            num_inferences = len(args[0]) if args else len(kwargs.get("images", []))
        else:
            # For single inference, increment by 1
            num_inferences = 1

        self.num_inferences += num_inferences
        self.average_latency = self.total_inference_time / self.num_inferences

        return result

    return wrapper


class BaseModel(ABC):
    def __init__(self, model_id: str, device: str, dtype: str):
        self.model_id = model_id
        self.device = device
        self.dtype = dtype
        self.num_inferences = 0
        self.total_inference_time = 0.0
        self.average_latency = 0.0

        logger.info(f"Model: {model_id}")
        logger.info(f"Device: {device}")
        logger.info(f"Dtype: {dtype}")

        dtype_map = {
            "float32": torch.float32,
            "float16": torch.float16,
            "bfloat16": torch.bfloat16,
            "auto": torch.float32,
        }
        if dtype not in dtype_map:
            raise ValueError("dtype must be one of 'float32', 'float16', or 'bfloat16'")
        self.dtype = dtype_map[dtype]

    @abstractmethod
    def load_model(self):
        """
        Load the model and any necessary components.

        This method should be implemented by subclasses to initialize
        their specific model architecture and weights.

        Raises:
            NotImplementedError: If the subclass doesn't implement this method
        """
        logger.info("Loading model...")
        raise NotImplementedError("Subclass must implement load_model()")

    @abstractmethod
    def infer(self, image: str, prompt: str):
        """
        Run inference on a single image.

        Args:
            image (str): Path or URL to the input image
            prompt (str): Text prompt for the inference

        Raises:
            NotImplementedError: If the subclass doesn't implement this method
        """
        logger.info("Running single inference...")
        raise NotImplementedError("Subclass must implement infer()")

    @abstractmethod
    def infer_batch(self, images: list[str], prompts: list[str]):
        """
        Run inference on a batch of images.

        Args:
            images (list[str]): List of image paths or URLs
            prompts (list[str]): List of text prompts for each image

        Raises:
            NotImplementedError: If the subclass doesn't implement this method
        """
        logger.info("Running batch inference...")
        raise NotImplementedError("Subclass must implement infer_batch()")

    def launch_gradio(self, **gradio_launch_kwargs):
        # Importing here to avoid circular import
        from .viz import launch_gradio

        launch_gradio(self, **gradio_launch_kwargs)

    def print_stats(self):
        console = Console()
        table = Table(title="Model Info", box=box.ROUNDED)
        table.add_column("Attribute", style="cyan")
        table.add_column("Value", style="magenta")

        table.add_row("Model ID", str(self.model_id))
        table.add_row("Device", str(self.device))
        table.add_row("Dtype", str(self.dtype))
        table.add_row("Number of Inferences", f"{self.num_inferences}")
        table.add_row("Total Inference Time (ms)", f"{self.total_inference_time:.4f}")
        table.add_row("Average Latency (ms)", f"{self.average_latency:.4f}")

        console.print(table)

    def __repr__(self) -> str:
        return (
            f"{self.__class__.__name__}("
            f"model_id='{self.model_id}', "
            f"device='{self.device}', "
            f"dtype='{self.dtype}', "
        )

    def parse_images(
        self,
        images: str | list[str],
    ) -> list[Image.Image]:
        """
        Preprocess one or more images from file paths or URLs.

        Loads and converts images to RGB format from either local file paths or URLs.
        Can handle both single image input or multiple images as a list.

        Args:
            images (Union[str, List[str]]): Either a single image path/URL as a string,
                or a list of image paths/URLs. Accepts both local file paths and HTTP(S) URLs.

        Returns:
            List[PIL.Image.Image]: List of processed PIL Image objects in RGB format.
        """

        if not isinstance(images, list):
            images = [images]

        parsed_images = []
        for image_path in images:
            if not isinstance(image_path, str):
                raise ValueError("Input must be a string (local path or URL)")

            if image_path.startswith(("http://", "https://")):
                image = Image.open(requests.get(image_path, stream=True).raw).convert(
                    "RGB"
                )
            else:
                # Assume it's a local path
                try:
                    image = Image.open(image_path).convert("RGB")
                except FileNotFoundError:
                    raise ValueError(f"Local file not found: {image_path}")

            parsed_images.append(image)

        return parsed_images



import timm

from ..model_registry import ModelInputOutput, register_model
from .timm_model import TimmModel


def load_timm_models():
    model_list = timm.list_models("*1k*", pretrained=True)

    for model_id in model_list:
        register_model(
            f"timm/{model_id}", "timm", ModelInputOutput.IMAGE_TO_CATEGORIES
        )(TimmModel)


load_timm_models()



from pathlib import Path

import pytest
import torch

import xinfer


@pytest.fixture
def model():
    return xinfer.create_model(
        "microsoft/Florence-2-base-ft", device="cpu", dtype="float32"
    )


@pytest.fixture
def test_image():
    return str(Path(__file__).parent.parent / "test_data" / "test_image_1.jpg")


def test_florence2_initialization(model):
    assert model.model_id == "microsoft/Florence-2-base-ft"
    assert model.device == "cpu"
    assert model.dtype == torch.float32


def test_florence2_inference(model, test_image):
    prompt = "<CAPTION>"
    result = model.infer(test_image, prompt)

    assert isinstance(result, str)
    assert len(result) > 0


def test_florence2_batch_inference(model, test_image):
    prompt = "<CAPTION>"
    result = model.infer_batch([test_image, test_image], [prompt, prompt])

    assert isinstance(result, list)
    assert len(result) == 2



import requests
import torch
from PIL import Image
from transformers import (
    AutoModelForVision2Seq,
    AutoProcessor,
)

from ..models import BaseModel, track_inference


class Vision2SeqModel(BaseModel):
    def __init__(
        self, model_id: str, device: str = "cpu", dtype: str = "float32", **kwargs
    ):
        super().__init__(model_id, device, dtype)
        self.load_model(**kwargs)

    def load_model(self, **kwargs):
        self.processor = AutoProcessor.from_pretrained(self.model_id, **kwargs)
        self.model = AutoModelForVision2Seq.from_pretrained(self.model_id, **kwargs).to(
            self.device, self.dtype
        )

        self.model = torch.compile(self.model, mode="max-autotune")
        self.model.eval()

    def preprocess(
        self,
        images: str | list[str],
        prompts: str | list[str],
    ):
        if not isinstance(images, list):
            images = [images]
        if not isinstance(prompts, list):
            prompts = [prompts]

        if len(images) != len(prompts):
            raise ValueError("The number of images and prompts must be the same")

        processed_images = []
        for image_path in images:
            if not isinstance(image_path, str):
                raise ValueError("Input must be a string (local path or URL)")

            if image_path.startswith(("http://", "https://")):
                image = Image.open(requests.get(image_path, stream=True).raw).convert(
                    "RGB"
                )
            else:
                # Assume it's a local path
                try:
                    image = Image.open(image_path).convert("RGB")
                except FileNotFoundError:
                    raise ValueError(f"Local file not found: {image_path}")

            processed_images.append(image)

        return self.processor(
            images=processed_images, text=prompts, return_tensors="pt"
        ).to(self.device, self.dtype)

    def predict(self, preprocessed_input, **generate_kwargs):
        with torch.inference_mode(), torch.amp.autocast(
            device_type=self.device, dtype=self.dtype
        ):
            return self.model.generate(**preprocessed_input, **generate_kwargs)

    def postprocess(self, predictions):
        outputs = self.processor.batch_decode(predictions, skip_special_tokens=True)
        return [output.replace("\n", "").strip() for output in outputs]

    @track_inference
    def infer(self, image, prompt, **generate_kwargs):
        preprocessed_input = self.preprocess(image, prompt)
        prediction = self.predict(preprocessed_input, **generate_kwargs)
        result = self.postprocess(prediction)[0]

        return result

    @track_inference
    def infer_batch(self, images, prompts, **generate_kwargs):
        preprocessed_input = self.preprocess(images, prompts)
        predictions = self.predict(preprocessed_input, **generate_kwargs)
        results = self.postprocess(predictions)

        return results



import torch
from transformers import AutoModelForCausalLM, AutoTokenizer

from ..model_registry import ModelInputOutput, register_model
from ..models import BaseModel, track_inference


@register_model(
    "vikhyatk/moondream2", "transformers", ModelInputOutput.IMAGE_TEXT_TO_TEXT
)
class Moondream(BaseModel):
    def __init__(
        self,
        model_id: str = "vikhyatk/moondream2",
        revision: str = "2024-08-26",
        device: str = "cpu",
        dtype: str = "float32",
    ):
        super().__init__(model_id, device, dtype)
        self.revision = revision
        self.load_model()

    def load_model(self):
        self.model = AutoModelForCausalLM.from_pretrained(
            self.model_id, trust_remote_code=True, revision=self.revision
        ).to(self.device, self.dtype)

        self.model = torch.compile(self.model, mode="max-autotune")
        self.model.eval()
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_id)

    @track_inference
    def infer(self, image: str, prompt: str = None, **generate_kwargs):
        image = self.parse_images(image)
        encoded_image = self.model.encode_image(image)
        output = self.model.answer_question(
            question=prompt,
            image_embeds=encoded_image,
            tokenizer=self.tokenizer,
            **generate_kwargs,
        )

        return output

    @track_inference
    def infer_batch(self, images: list[str], prompts: list[str], **generate_kwargs):
        images = self.parse_images(images)
        prompts = [prompt for prompt in prompts]

        outputs = self.model.batch_answer(
            images, prompts, self.tokenizer, **generate_kwargs
        )

        return outputs



import torch

from ..model_registry import ModelInputOutput, register_model
from .blip2 import BLIP2


@register_model(
    "sashakunitsyn/vlrm-blip2-opt-2.7b",
    "transformers",
    ModelInputOutput.IMAGE_TEXT_TO_TEXT,
)
class VLRMBlip2(BLIP2):
    def __init__(self, model_name: str = "sashakunitsyn/vlrm-blip2-opt-2.7b", **kwargs):
        super().__init__(model_name, **kwargs)
        self.load_vlrm_weights()

    def load_vlrm_weights(self):
        from huggingface_hub import hf_hub_download

        finetuned_weights_state_dict = torch.load(
            hf_hub_download(
                repo_id="sashakunitsyn/vlrm-blip2-opt-2.7b",
                filename="vlrm-blip2-opt-2.7b.pt",
            )
        )
        self.model.load_state_dict(finetuned_weights_state_dict, strict=False)



import importlib


def soft_import(name: str):
    try:
        importlib.import_module(name)
        return True
    except ModuleNotFoundError as e:
        if str(e) != f"No module named '{name}'":
            raise e
        return False


# Check for available modules
timm_available = soft_import("timm")
transformers_available = soft_import("transformers")
ultralytics_available = soft_import("ultralytics")
vllm_available = soft_import("vllm")
ollama_available = soft_import("ollama")


# Create placeholder classes
def create_placeholder(name):
    return type(name, (), {})


# Import or create placeholder classes so that we can still use the class types
if timm_available:
    from .timm.timm_model import TimmModel
else:
    TimmModel = create_placeholder("TimmModel")

if transformers_available:
    from .transformers.vision2seq import Vision2SeqModel
else:
    Vision2SeqModel = create_placeholder("Vision2SeqModel")

if ultralytics_available:
    from .ultralytics.ultralytics_model import UltralyticsModel
else:
    UltralyticsModel = create_placeholder("UltralyticsModel")



import torch
from transformers import AutoProcessor, MllamaForConditionalGeneration

from ..model_registry import ModelInputOutput, register_model
from ..models import BaseModel, track_inference


@register_model(
    "meta-llama/Llama-3.2-11B-Vision-Instruct",
    "transformers",
    ModelInputOutput.IMAGE_TEXT_TO_TEXT,
)
@register_model(
    "meta-llama/Llama-3.2-90B-Vision-Instruct",
    "transformers",
    ModelInputOutput.IMAGE_TEXT_TO_TEXT,
)
class Llama32VisionInstruct(BaseModel):
    def __init__(
        self, model_id: str, device: str = "cpu", dtype: str = "float32", **kwargs
    ):
        super().__init__(model_id, device, dtype)
        self.load_model(**kwargs)

    def load_model(self, **kwargs):
        # Use device_map="auto" to automatically distribute model across available devices
        self.model = MllamaForConditionalGeneration.from_pretrained(
            self.model_id, torch_dtype=self.dtype, device_map="auto", **kwargs
        )
        self.model.eval()
        self.processor = AutoProcessor.from_pretrained(self.model_id)

    @track_inference
    def infer(self, image: str, prompt: str, **generate_kwargs) -> str:
        image = super().parse_images(image)

        # Create messages format and apply chat template
        messages = [
            {
                "role": "user",
                "content": [{"type": "image"}, {"type": "text", "text": prompt}],
            }
        ]
        input_text = self.processor.apply_chat_template(
            messages, add_generation_prompt=True
        )

        # Process inputs without adding special tokens
        inputs = self.processor(
            image, input_text, add_special_tokens=False, return_tensors="pt"
        ).to(self.model.device)

        with torch.inference_mode():
            output = self.model.generate(**inputs, **generate_kwargs)

        decoded = self.processor.decode(output[0], skip_special_tokens=True)
        # Remove the prompt and assistant marker
        if "assistant" in decoded:
            decoded = decoded.split("assistant")[-1]
        return decoded.strip()

    def infer_batch(self, images: list[str], prompts: list[str], **generate_kwargs):
        images = super().parse_images(images)

        # Create batch messages and apply chat template
        messages_list = [
            [
                {
                    "role": "user",
                    "content": [{"type": "image"}, {"type": "text", "text": prompt}],
                }
            ]
            for prompt in prompts
        ]
        input_texts = [
            self.processor.apply_chat_template(msgs, add_generation_prompt=True)
            for msgs in messages_list
        ]

        inputs = self.processor(
            images, input_texts, add_special_tokens=False, return_tensors="pt"
        ).to(self.model.device)

        with torch.inference_mode():
            outputs = self.model.generate(**inputs, **generate_kwargs)

        decoded = self.processor.batch_decode(outputs, skip_special_tokens=True)
        # Remove the prompt and assistant marker for each response
        return [d.split("assistant")[-1].strip() for d in decoded]


@register_model(
    "meta-llama/Llama-3.2-11B-Vision",
    "transformers",
    ModelInputOutput.IMAGE_TEXT_TO_TEXT,
)
@register_model(
    "meta-llama/Llama-3.2-90B-Vision",
    "transformers",
    ModelInputOutput.IMAGE_TEXT_TO_TEXT,
)
class Llama32Vision(Llama32VisionInstruct):
    @track_inference
    def infer(self, image: str, prompt: str, **generate_kwargs) -> str:
        image = super().parse_images(image)

        # Format prompt for base vision model
        input_text = f"<|image|><|begin_of_text|>{prompt}"

        # Process inputs without adding special tokens
        inputs = self.processor(image, input_text, return_tensors="pt").to(
            self.model.device
        )

        with torch.inference_mode():
            output = self.model.generate(**inputs, **generate_kwargs)

        return self.processor.decode(output[0], skip_special_tokens=True)

    def infer_batch(self, images: list[str], prompts: list[str], **generate_kwargs):
        images = super().parse_images(images)

        # Format prompts for base vision model
        input_texts = [f"<|image|><|begin_of_text|>{prompt}" for prompt in prompts]

        inputs = self.processor(images, input_texts, return_tensors="pt").to(
            self.model.device
        )

        with torch.inference_mode():
            outputs = self.model.generate(**inputs, **generate_kwargs)

        return self.processor.batch_decode(outputs, skip_special_tokens=True)



from pathlib import Path

import pytest
import torch

import xinfer


@pytest.fixture
def model():
    return xinfer.create_model("vikhyatk/moondream2", device="cpu", dtype="float32")


@pytest.fixture
def test_image():
    return str(Path(__file__).parent.parent / "test_data" / "test_image_1.jpg")


def test_moondream_initialization(model):
    assert model.model_id == "vikhyatk/moondream2"
    assert model.device == "cpu"
    assert model.dtype == torch.float32


def test_moondream_inference(model, test_image):
    prompt = "Caption this image."
    result = model.infer(test_image, prompt)

    assert isinstance(result, str)
    assert len(result) > 0


def test_moondream_batch_inference(model, test_image):
    prompt = "Caption this image."
    result = model.infer_batch([test_image, test_image], [prompt, prompt])

    assert isinstance(result, list)
    assert len(result) == 2



import json

import gradio as gr

from .core import create_model
from .model_registry import ModelInputOutput, model_registry
from .models import BaseModel


def launch_gradio(model: BaseModel, **gradio_launch_kwargs):
    model_info = model_registry.get_model_info(model.model_id)

    def infer(image, prompt=None):
        try:
            if prompt is not None:
                result = model.infer(image, prompt)
            else:
                result = model.infer(image)

            # Convert result to string if it's not already
            if isinstance(result, str):
                return result
            else:
                return json.dumps(result, indent=2)
        except Exception as e:
            return f"Error during inference: {str(e)}"

    if model_info.input_output == ModelInputOutput.IMAGE_TEXT_TO_TEXT:
        iface = gr.Interface(
            fn=infer,
            inputs=[gr.Image(type="filepath"), gr.Textbox(label="Prompt")],
            outputs=gr.Textbox(label="Result", lines=10),
            title=f"Inference with {model.model_id}",
            description="Upload an image and provide a prompt to generate a description.",
        )

    elif model_info.input_output == ModelInputOutput.IMAGE_TO_BOXES:
        iface = gr.Interface(
            fn=infer,
            inputs=gr.Image(type="filepath"),
            outputs=gr.Textbox(label="Result", lines=10),
            title=f"Object Detection with {model.model_id}",
            description="Upload an image to detect objects.",
        )

    elif model_info.input_output == ModelInputOutput.IMAGE_TO_CATEGORIES:
        iface = gr.Interface(
            fn=infer,
            inputs=gr.Image(type="filepath"),
            outputs=gr.Textbox(label="Result", lines=10),
            title=f"Image Classification with {model.model_id}",
            description="Upload an image to classify.",
        )

    # The default height of Gradio is too small for view in jupyter notebooks
    if "height" not in gradio_launch_kwargs:
        gradio_launch_kwargs["height"] = 1000

    iface.launch(**gradio_launch_kwargs)


def launch_gradio_demo():
    """
    Launch an interactive demo with a dropdown to select a model from all supported models,
    and a button to run inference.
    """
    available_models = [model.id for model in model_registry.list_models()]

    # Add example image URLs
    example_images = [
        "https://raw.githubusercontent.com/dnth/x.infer/refs/heads/main/assets/demo/000b9c365c9e307a.jpg",
        "https://raw.githubusercontent.com/dnth/x.infer/refs/heads/main/assets/demo/00aa2580828a9009.jpg",
        "https://raw.githubusercontent.com/dnth/x.infer/refs/heads/main/assets/demo/0a6ee446579d2885.jpg",
    ]

    def load_model_and_infer(model_id, image, prompt, device, dtype):
        model = create_model(model_id, device=device, dtype=dtype)
        model_info = model_registry.get_model_info(model_id)

        try:
            if model_info.input_output == ModelInputOutput.IMAGE_TEXT_TO_TEXT:
                result = model.infer(image, prompt)
            elif model_info.input_output in [
                ModelInputOutput.IMAGE_TO_BOXES,
                ModelInputOutput.IMAGE_TO_CATEGORIES,
            ]:
                result = model.infer(image)
            else:
                return "Unsupported model type"

            # Convert result to string if it's not already
            if isinstance(result, str):
                return result
            else:
                return json.dumps(result, indent=2)
        except Exception as e:
            return f"Error during inference: {str(e)}"

    with gr.Blocks() as demo:
        gr.Markdown("# x.infer Gradio Demo")

        with gr.Row():
            # Left column: Input controls
            with gr.Column(scale=1):
                model_dropdown = gr.Dropdown(
                    choices=available_models, label="Select a model"
                )
                with gr.Row():
                    device_dropdown = gr.Dropdown(
                        choices=["cuda", "cpu"], label="Device", value="cuda"
                    )
                    dtype_dropdown = gr.Dropdown(
                        choices=["float32", "float16", "bfloat16"],
                        label="Dtype",
                        value="float16",
                    )
                prompt_input = gr.Textbox(
                    label="Prompt (for image-text to text models)", visible=False
                )
                run_button = gr.Button("Run Inference", variant="primary")

            # Right column: Image input
            with gr.Column(scale=1):
                image_input = gr.Image(type="filepath", label="Input Image", height=400)

        # Results section
        output = gr.Textbox(label="Result", lines=5)

        # Add examples
        gr.Examples(examples=example_images, inputs=image_input, label="Example Images")

        def update_prompt_visibility(model_id):
            model_info = model_registry.get_model_info(model_id)
            return gr.update(
                visible=model_info.input_output == ModelInputOutput.IMAGE_TEXT_TO_TEXT
            )

        model_dropdown.change(
            update_prompt_visibility, inputs=[model_dropdown], outputs=[prompt_input]
        )

        run_button.click(
            load_model_and_infer,
            inputs=[
                model_dropdown,
                image_input,
                prompt_input,
                device_dropdown,
                dtype_dropdown,
            ],
            outputs=[output],
        )

    demo.launch(height=1000)



from vllm import LLM, SamplingParams

from ..model_registry import ModelInputOutput, register_model
from ..models import BaseModel, track_inference


@register_model(
    "vllm/allenai/Molmo-72B-0924", "vllm", ModelInputOutput.IMAGE_TEXT_TO_TEXT
)
@register_model(
    "vllm/allenai/Molmo-7B-O-0924", "vllm", ModelInputOutput.IMAGE_TEXT_TO_TEXT
)
@register_model(
    "vllm/allenai/Molmo-7B-D-0924", "vllm", ModelInputOutput.IMAGE_TEXT_TO_TEXT
)
class Molmo(BaseModel):
    def __init__(
        self,
        model_id: str,
        device: str = "cpu",
        dtype: str = "float32",
        **kwargs,
    ):
        super().__init__(model_id, device, dtype)
        self.load_model(**kwargs)

    def load_model(self, **kwargs):
        self.model = LLM(
            model=self.model_id,
            trust_remote_code=True,
            dtype=self.dtype,
            **kwargs,
        )

    @track_inference
    def infer_batch(self, images: list[str], prompts: list[str], **sampling_kwargs):
        images = self.parse_images(images)

        sampling_params = SamplingParams(**sampling_kwargs)
        batch_inputs = [
            {
                "prompt": f"USER: <image>\n{prompt}\nASSISTANT:",
                "multi_modal_data": {"image": image},
            }
            for image, prompt in zip(images, prompts)
        ]

        results = self.model.generate(batch_inputs, sampling_params)

        return [output.outputs[0].text.strip() for output in results]

    @track_inference
    def infer(self, image: str, prompt: str, **sampling_kwargs):
        image = self.parse_images(image)

        inputs = {
            "prompt": prompt,
            "multi_modal_data": {"image": image},
        }

        sampling_params = SamplingParams(**sampling_kwargs)
        outputs = self.model.generate(inputs, sampling_params)
        generated_text = outputs[0].outputs[0].text.strip()
        return generated_text



from loguru import logger
from rich.console import Console
from rich.table import Table

from .model_registry import model_registry
from .utils import TimmModel, UltralyticsModel, Vision2SeqModel


def create_model(model: str | TimmModel | Vision2SeqModel | UltralyticsModel, **kwargs):
    """
    Create a model instance.

    Parameters
    ----------
    model : str | TimmModel | Vision2SeqModel | UltralyticsModel
        The model to create.
        TIMM, Vision2Seq, and Ultralytics models type here is to support user passing in the models directly.
        This is useful for models not registered in the model registry.

        Eg:
        ```python
        model = UltralyticsModel("yolov5n6u")
        model = xinfer.create_model(model)
        ```
    """
    if isinstance(model, (TimmModel, Vision2SeqModel, UltralyticsModel)):
        return model
    return model_registry.get_model(model, **kwargs)


def list_models(search: str = None, limit: int = 20, interactive: bool = False):
    import pandas as pd

    rows = []
    for model_info in model_registry.list_models():
        if search is None or search.lower() in model_info.id.lower():
            rows.append(
                {
                    "Implementation": model_info.implementation,
                    "Model ID": model_info.id,
                    "Input --> Output": model_info.input_output.value,
                }
            )

    if not rows:
        logger.warning(
            "No models found matching the criteria.\n"
            "Perhaps install the relevant dependencies? For example, `pip install xinfer[timm]`"
        )
        return

    if interactive:
        from itables import init_notebook_mode

        logger.info(
            "Showing interactive table in Jupyter Notebook. Type in the search bar to filter the models."
        )

        init_notebook_mode(all_interactive=True)
        return pd.DataFrame(rows)

    if len(rows) > limit:
        rows = rows[:limit]
        rows.append(
            {"Implementation": "...", "Model ID": "...", "Input --> Output": "..."}
        )
        rows.append(
            {"Implementation": "...", "Model ID": "...", "Input --> Output": "..."}
        )

        logger.info(
            f"Showing only top {limit} models. Change the `limit` parameter to see more."
        )

    console = Console()
    table = Table(title="Available Models")
    table.add_column("Implementation", style="cyan")
    table.add_column("Model ID", style="magenta")
    table.add_column("Input --> Output", style="green")

    for row in rows:
        table.add_row(row["Implementation"], row["Model ID"], row["Input --> Output"])

    console.print(table)



from ..model_registry import ModelInputOutput, register_model
from .ultralytics_model import UltralyticsModel


@register_model("ultralytics/yolov11n", "ultralytics", ModelInputOutput.IMAGE_TO_BOXES)
@register_model("ultralytics/yolov11s", "ultralytics", ModelInputOutput.IMAGE_TO_BOXES)
@register_model("ultralytics/yolov11m", "ultralytics", ModelInputOutput.IMAGE_TO_BOXES)
@register_model("ultralytics/yolov11l", "ultralytics", ModelInputOutput.IMAGE_TO_BOXES)
class YOLOv11(UltralyticsModel):
    def __init__(self, model_id: str, **kwargs):
        model_id = model_id.replace("v", "")
        super().__init__(model_id, **kwargs)



import torch
import torchvision.transforms.functional as TVF
from PIL import Image
from transformers import AutoTokenizer, LlavaForConditionalGeneration

from ..model_registry import ModelInputOutput, register_model
from ..models import BaseModel, track_inference


@register_model(
    "fancyfeast/llama-joycaption-alpha-two-hf-llava",
    "transformers",
    ModelInputOutput.IMAGE_TEXT_TO_TEXT,
)
class JoyCaption(BaseModel):
    def __init__(
        self, model_id: str, device: str = "cpu", dtype: str = "float32", **kwargs
    ):
        super().__init__(model_id, device, dtype)
        self.load_model(**kwargs)

    def load_model(self, **kwargs):
        self.tokenizer = AutoTokenizer.from_pretrained(self.model_id, use_fast=True)
        self.model = LlavaForConditionalGeneration.from_pretrained(
            self.model_id, **kwargs
        ).to(self.device, self.dtype)
        self.model.eval()

    def preprocess(self, image: str, prompt: str = None):
        image = Image.open(image).convert("RGB")
        image = TVF.resize(image, (384, 384), Image.LANCZOS)
        image = TVF.pil_to_tensor(image)

        image = image / 255.0
        image = TVF.normalize(image, [0.5], [0.5])
        image = image.to(self.dtype).unsqueeze(0)

        # Build the conversation
        convo = [
            {
                "role": "system",
                "content": "You are a helpful image captioner.",
            },
            {
                "role": "user",
                "content": prompt,
            },
        ]

        # Format the conversation
        convo_string = self.tokenizer.apply_chat_template(
            convo, tokenize=False, add_generation_prompt=True
        )

        # Tokenize the conversation
        convo_tokens = self.tokenizer.encode(
            convo_string, add_special_tokens=False, truncation=False
        )

        # Repeat the image tokens
        input_tokens = []
        for token in convo_tokens:
            if token == self.model.config.image_token_index:
                input_tokens.extend(
                    [self.model.config.image_token_index]
                    * self.model.config.image_seq_length
                )
            else:
                input_tokens.append(token)

        input_ids = torch.tensor(input_tokens, dtype=torch.long).unsqueeze(0)
        attention_mask = torch.ones_like(input_ids)

        return image, input_ids, attention_mask

    @track_inference
    def infer(self, image: str, prompt: str = None, **generate_kwargs):
        with torch.inference_mode(), torch.amp.autocast(
            device_type=self.device, dtype=self.dtype
        ):
            image, input_ids, attention_mask = self.preprocess(image, prompt)

            if "max_new_tokens" not in generate_kwargs:
                generate_kwargs["max_new_tokens"] = 300

            generate_ids = self.model.generate(
                input_ids=input_ids.to(self.device),
                pixel_values=image.to(self.device),
                attention_mask=attention_mask.to(self.device),
                do_sample=True,
                suppress_tokens=None,
                use_cache=True,
                **generate_kwargs,
            )[0]

        # Trim off the prompt
        generate_ids = generate_ids[input_ids.shape[1] :]

        # Decode the caption
        caption = self.tokenizer.decode(
            generate_ids,
            skip_special_tokens=True,
            clean_up_tokenization_spaces=False,
        )
        caption = caption.strip()

        return caption

    def preprocess_batch(self, images: list[str], prompts: list[str]):
        processed_images = []
        input_ids_list = []
        attention_masks = []

        for image, prompt in zip(images, prompts):
            image = Image.open(image).convert("RGB")
            image = TVF.resize(image, (384, 384), Image.LANCZOS)
            image = TVF.pil_to_tensor(image)
            image = image / 255.0
            image = TVF.normalize(image, [0.5], [0.5])
            image = image.to(self.dtype)
            processed_images.append(image)

            convo = [
                {"role": "system", "content": "You are a helpful image captioner."},
                {"role": "user", "content": prompt},
            ]
            convo_string = self.tokenizer.apply_chat_template(
                convo, tokenize=False, add_generation_prompt=True
            )
            convo_tokens = self.tokenizer.encode(
                convo_string, add_special_tokens=False, truncation=False
            )

            input_tokens = []
            for token in convo_tokens:
                if token == self.model.config.image_token_index:
                    input_tokens.extend(
                        [self.model.config.image_token_index]
                        * self.model.config.image_seq_length
                    )
                else:
                    input_tokens.append(token)

            input_ids = torch.tensor(input_tokens, dtype=torch.long)
            input_ids_list.append(input_ids)

        # Pad input_ids to the same length
        max_length = max(ids.size(0) for ids in input_ids_list)
        padded_input_ids = torch.zeros(
            (len(input_ids_list), max_length), dtype=torch.long
        )
        for i, ids in enumerate(input_ids_list):
            padded_input_ids[i, : ids.size(0)] = ids
            attention_masks.append(torch.ones(max_length, dtype=torch.long))
            attention_masks[-1][ids.size(0) :] = 0

        images_tensor = torch.stack(processed_images)
        attention_mask_tensor = torch.stack(attention_masks)

        return images_tensor, padded_input_ids, attention_mask_tensor

    @track_inference
    def infer_batch(self, images: list[str], prompts: list[str], **generate_kwargs):
        with torch.inference_mode(), torch.amp.autocast(
            device_type=self.device, dtype=self.dtype
        ):
            images, input_ids, attention_mask = self.preprocess_batch(images, prompts)

            if "max_new_tokens" not in generate_kwargs:
                generate_kwargs["max_new_tokens"] = 300

            generate_ids = self.model.generate(
                input_ids=input_ids.to(self.device),
                pixel_values=images.to(self.device),
                attention_mask=attention_mask.to(self.device),
                do_sample=True,
                suppress_tokens=None,
                use_cache=True,
                **generate_kwargs,
            )

        captions = []
        for i, gen_ids in enumerate(generate_ids):
            # Trim off the prompt
            gen_ids = gen_ids[input_ids.shape[1] :]

            # Decode the caption
            caption = self.tokenizer.decode(
                gen_ids,
                skip_special_tokens=True,
                clean_up_tokenization_spaces=False,
            )
            captions.append(caption.strip())

        return captions



from typing import Dict, List

import torch
from ultralytics import YOLO

from ..models import BaseModel, track_inference


class UltralyticsModel(BaseModel):
    def __init__(
        self, model_id: str, device: str = "cpu", dtype: str = "float32", **kwargs
    ):
        super().__init__(model_id, device, dtype)
        self.load_model(**kwargs)

    def load_model(self, **kwargs):
        self.model = YOLO(self.model_id, **kwargs)

    @track_inference
    def infer_batch(self, images: str | List[str], **kwargs) -> List[List[Dict]]:
        half = self.dtype == torch.float16
        results = self.model.predict(images, device=self.device, half=half, **kwargs)

        batch_results = []
        for result in results:
            coco_format_results = []
            boxes = result.boxes
            for box in boxes:
                x1, y1, x2, y2 = box.xyxy[0].tolist()
                width = x2 - x1
                height = y2 - y1
                coco_format_results.append(
                    {
                        "bbox": [x1, y1, width, height],
                        "category_id": int(box.cls),
                        "score": float(box.conf),
                        "class_name": result.names[int(box.cls)],
                    }
                )
            batch_results.append(coco_format_results)
        return batch_results

    @track_inference
    def infer(self, image: str, **kwargs) -> List[List[Dict]]:
        results = self.infer_batch([image], **kwargs)
        return results[0]



from dataclasses import dataclass
from enum import Enum
from typing import Dict, List, Type

from .models import BaseModel


class ModelInputOutput(Enum):
    IMAGE_TO_TEXT = "image --> text"
    IMAGE_TEXT_TO_TEXT = "image-text --> text"
    TEXT_TO_TEXT = "text --> text"
    IMAGE_TO_BOXES = "image --> boxes"
    IMAGE_TO_CATEGORIES = "image --> categories"


@dataclass
class ModelInfo:
    id: str
    implementation: str
    input_output: ModelInputOutput


class ModelRegistry:
    def __init__(self):
        self._models: Dict[str, ModelInfo] = {}

    def register(self, model_info: ModelInfo, model_class: Type[BaseModel]):
        if model_info.id in self._models:
            raise ValueError(
                f"Model {model_info.id} already registered. Pick another id."
            )
        self._models[model_info.id] = (model_info, model_class)

    def get_model(self, model_id: str, **kwargs) -> BaseModel:
        model_info, model_class = self._models.get(model_id, (None, None))
        if model_class is None:
            raise ValueError(f"Unsupported model: {model_id}")
        return model_class(model_id, **kwargs)

    def list_models(self) -> List[ModelInfo]:
        return [model_info for model_info, _ in self._models.values()]

    def get_model_info(self, name: str) -> ModelInfo:
        model_info, _ = self._models.get(name, (None, None))
        if model_info is None:
            raise ValueError(f"Unsupported model: {name}")
        return model_info


# Create a global instance of the registry
model_registry = ModelRegistry()


def register_model(model_id: str, implementation: str, input_output: ModelInputOutput):
    def decorator(cls: Type[BaseModel]):
        model_registry.register(ModelInfo(model_id, implementation, input_output), cls)
        return cls

    return decorator



from io import BytesIO
from typing import Dict, List

import requests
import timm
import torch
from PIL import Image

from ..models import BaseModel, track_inference
from .imagenet1k_classes import IMAGENET2012_CLASSES


class TimmModel(BaseModel):
    def __init__(
        self, model_id: str, device: str = "cpu", dtype: str = "float32", **kwargs
    ):
        super().__init__(model_id, device, dtype)
        self.load_model(**kwargs)

    def load_model(self, **kwargs):
        self.model = timm.create_model(self.model_id, pretrained=True, **kwargs).to(
            self.device, self.dtype
        )
        # self.model = torch.compile(self.model, mode="max-autotune")
        self.model.eval()

    def preprocess(self, images: str | list[str]):
        if not isinstance(images, list):
            images = [images]

        processed_images = []
        for image_path in images:
            if not isinstance(image_path, str):
                raise ValueError("Input must be a string (local path or URL)")

            if image_path.startswith(("http://", "https://")):
                response = requests.get(image_path)
                img = Image.open(BytesIO(response.content))
            else:
                img = Image.open(image_path)

            processed_images.append(img)

        data_config = timm.data.resolve_model_data_config(self.model)
        transforms = timm.data.create_transform(**data_config, is_training=False)

        images_tensor = torch.stack([transforms(img) for img in processed_images])
        images_tensor = images_tensor.to(device=self.device, dtype=self.dtype)

        return images_tensor

    @track_inference
    def infer(self, image: str, top_k: int = 5) -> List[Dict]:
        img = self.preprocess(image)

        with torch.inference_mode(), torch.amp.autocast(
            device_type=self.device, dtype=self.dtype
        ):
            output = self.model(img)

        topk_probabilities, topk_class_indices = torch.topk(
            output.softmax(dim=1) * 100, k=top_k
        )

        im_classes = list(IMAGENET2012_CLASSES.values())
        class_names = [im_classes[i] for i in topk_class_indices[0]]

        return [
            {"class": class_name, "id": int(class_idx), "confidence": float(prob)}
            for class_name, class_idx, prob in zip(
                class_names, topk_class_indices[0], topk_probabilities[0]
            )
        ]

    @track_inference
    def infer_batch(self, images: List[str], top_k: int = 5) -> List[List[Dict]]:
        images = self.preprocess(images)

        with torch.inference_mode(), torch.amp.autocast(
            device_type=self.device, dtype=self.dtype
        ):
            output = self.model(images)

        topk_probabilities, topk_class_indices = torch.topk(
            output.softmax(dim=1) * 100, k=top_k
        )

        im_classes = list(IMAGENET2012_CLASSES.values())

        results = []
        for i in range(len(images)):
            class_names = [im_classes[idx] for idx in topk_class_indices[i]]
            image_results = [
                {"class": class_name, "id": int(class_idx), "confidence": float(prob)}
                for class_name, class_idx, prob in zip(
                    class_names, topk_class_indices[i], topk_probabilities[i]
                )
            ]
            results.append(image_results)

        return results



from .blip2 import BLIP2
from .florence2 import Florence2
from .joycaption import JoyCaption
from .llama32 import Llama32Vision, Llama32VisionInstruct
from .moondream import Moondream
from .vision2seq import Vision2SeqModel
from .vlrm_blip2 import VLRMBlip2



import ollama

from ..model_registry import ModelInputOutput, register_model
from ..models import BaseModel


@register_model("ollama/llava-phi3", "ollama", ModelInputOutput.IMAGE_TEXT_TO_TEXT)
class LLaVAPhi3(BaseModel):
    def __init__(
        self,
        model_id: str = "llava-phi3",
        device: str = "auto",
        dtype: str = "auto",
    ):
        super().__init__(model_id, device, dtype)
        self.load_model()

    def load_model(self):
        ollama.pull(self.model_id.replace("ollama/", ""))

    def infer_batch(self, image: str, prompt: str):
        raise NotImplementedError("Ollama models do not support batch inference")

    def infer(self, image: str, prompt: str) -> str:
        res = ollama.chat(
            model="llava-phi3",
            messages=[
                {
                    "role": "user",
                    "content": prompt,
                    "images": [image],
                }
            ],
        )
        return res["message"]["content"]



from ..model_registry import ModelInputOutput, register_model
from .vision2seq import Vision2SeqModel


@register_model(
    "Salesforce/blip2-opt-2.7b", "transformers", ModelInputOutput.IMAGE_TEXT_TO_TEXT
)
@register_model(
    "Salesforce/blip2-opt-6.7b", "transformers", ModelInputOutput.IMAGE_TEXT_TO_TEXT
)
@register_model(
    "Salesforce/blip2-flan-t5-xxl", "transformers", ModelInputOutput.IMAGE_TEXT_TO_TEXT
)
@register_model(
    "Salesforce/blip2-opt-6.7b-coco",
    "transformers",
    ModelInputOutput.IMAGE_TEXT_TO_TEXT,
)
class BLIP2(Vision2SeqModel):
    def __init__(self, model_id: str, **kwargs):
        super().__init__(model_id, **kwargs)



from pathlib import Path

import pytest
import torch

import xinfer


@pytest.fixture
def model():
    return xinfer.create_model("ultralytics/yolov8n", device="cpu", dtype="float32")


@pytest.fixture
def test_image():
    return str(Path(__file__).parent.parent / "test_data" / "test_image_1.jpg")


def test_ultralytics_initialization(model):
    assert model.model_id == "ultralytics/yolov8n"
    assert model.device == "cpu"
    assert model.dtype == torch.float32


def test_ultralytics_inference(model, test_image):
    # Test if there is boxes and scores in the first element of the result
    result = model.infer(test_image)[0]

    assert isinstance(result, dict)
    assert "bbox" in result
    assert "score" in result
    assert "class_name" in result
    assert "category_id" in result

    # Test bbox format and values
    assert isinstance(result["bbox"], list)
    assert len(result["bbox"]) == 4  # [x, y, width, height]
    assert all(isinstance(coord, float) for coord in result["bbox"])
    assert all(coord >= 0 for coord in result["bbox"])


def test_ultralytics_batch_inference(model, test_image):
    result = model.infer_batch([test_image, test_image])

    assert isinstance(result, list)
    assert len(result) == 2

    # Verify structure of each batch result
    for batch_result in result:
        assert isinstance(batch_result, list)
        # Check each detection in the batch
        for detection in batch_result:
            assert isinstance(detection, dict)
            assert "bbox" in detection
            assert "score" in detection
            assert "class_name" in detection
            assert "category_id" in detection

            # Verify data types and value ranges
            assert isinstance(detection["bbox"], list)
            assert len(detection["bbox"]) == 4  # [x, y, width, height]
            assert isinstance(detection["score"], float)
            assert 0 <= detection["score"] <= 1  # Score should be between 0 and 1
            assert isinstance(detection["class_name"], str)
            assert isinstance(detection["category_id"], int)
